const request = require('request');
const hash = require('wwl-hash');
const {promisify} = require('util');
const _glob = require('glob');
const glob = promisify(_glob);
var colors = require("colors");
const fs = require('fs');
const path = require('path')
const Minimatch = require("minimatch").Minimatch;

var isShowLog = false;
var consoleLog = function (color, type) {
    console.log(color);
};
const printLog = function (color, type) {
    if (!isShowLog) return;
    consoleLog(color, type);
}
const logger = {
    log: (txt) => {
        printLog(txt.yellow, 'log')
    },
    err: (txt) => {
        printLog(txt.red, 'error')
    },
    success: (txt) => {
        printLog(txt.green, 'success')
    }
}

//二维数组
const flatArray = function (arr) {
    return arr.reduce((pre, cur) => pre.concat(cur), []);
}

const sepReg = /\\|\//g;

const req = function (url, method, json, form) {
    return new Promise((resolve, reject) => {
        request({
            url,
            baseUrl: 'http://upload.ws.netease.com/',
            method,
            json,
            formData: form
        }, (error, response, body) => {
            if (error) reject(error);
            else {
                resolve({response, body});
            }
        })
    })
}

const postJson = function (url, json) {
    return req(url, 'POST', json, null);
}
const postForm = function (url, form) {
    return req(url, 'POST', null, form);
}


const auth = function (username, password) {
    logger.log('发起认证用户请求');
    return postJson('/fileupload/auth', {username, password}).then(
        ({response, body}) => {
            if (typeof body !== 'object' || body.resultcode !== 8200) {
                logger.err(`认证用户失败: ${JSON.stringify(body)}`);
                return Promise.reject({response, body});
            }
            else {
                logger.success('认证用户成功');
                return body.data;
            }
        },
        (err) => {
            logger.err('认证用户失败: 请求出错。');
            return Promise.reject(err);
        }
    )
}

var authPromise;
var getValidAuth = function (user, pwd) {
    if (!authPromise) {
        return authPromise = auth(user, pwd);
    } else {
        return authPromise = authPromise.then(session => session, err => {
            return auth(user, pwd)
        });
    }
}


const getFilesPath = function (pattern, opt = {}) {

    if (Array.isArray(pattern)) {
        return Promise.all(pattern.map(p => getFilesPath(p, opt))).then(results => {
            return flatArray(results);
        })
    }


    return glob(pattern, Object.assign(opt, {absolute: true})).then((files) => {
        return files.filter(file => {
            return fs.statSync(file).isFile()
        }).map(file => {
            //在windows上，分隔符为 \
            //但是glob返回的分隔符还是 /
            return file.replace(sepReg, path.sep);
        })
    });
}

const validDir = function (dirPath) {
    return dirPath
    // if (path.extname(dirPath) ) {
    //     dirPath = path.dirname(dirPath);
    //     logger.log(`目录路径中包含疑似文件名: ${dirPath}`)
    // }
    // return dirPath;
}

const uploadFile = function (session, filePath, onlinePath = '', baseFilePath, strict) {

    var fileStream = fs.createReadStream(filePath);
    var md5 = hash.hashFileSync(filePath, 'md5');
    var fileName = path.basename(filePath);

    var cdnPath;

    //兼容onlinePath为function
    if (typeof onlinePath === 'function') {
        cdnPath = onlinePath(filePath) || '';
    }
    else {
        cdnPath = onlinePath;
    }

    //兼容: 当onlinePath不是string的情况
    if (cdnPath) {
        cdnPath = (cdnPath + '').trim();
    }

    //cdnPath不存在或者是空字符串
    if (!cdnPath) {
        cdnPath = 'f2e/test';
    }
    else {
        //1.确保cdnPath是个路径， 比如xx/xx.js会被转化为 xx/
        //2.去掉收尾空格，比如 /a/b/ 会被转化为 a/b

        //检查是否是路径
        cdnPath = validDir(cdnPath);

        //去掉首尾的 \ 和 /
        cdnPath = cdnPath.replace(/^[\\/]|[\\/]$/g, '');

    }

    //保证文件的目录结构
    if (strict) {
        var dir = path.dirname(filePath); //结尾不包含 /
        if (dir.indexOf(baseFilePath) > -1) {
            cdnPath = path.join(cdnPath, path.relative(baseFilePath, dir));
        }
    }

    //在windows上，通过path.xxx处理后的路径，路径分隔符会变为本地系统的分隔符。
    cdnPath = cdnPath.replace(sepReg, '\/');

    var startTime = Date.now();

    //file md5 filename session path
    return postForm('/fileupload/add', {
        file: fileStream,
        md5: md5,
        filename: fileName,
        session: session,
        path: cdnPath
    }).then(
        ({response, body}) => {
            //返回err
            try {
                if (typeof body === 'string') body = JSON.parse(body);
            }
            catch (e) {
                logger.err(`[×] ${filePath} : JSON.parse(body)出错, ${body}`);

                return body;
            }

            var err;

            if (body.resultcode === 200) {
                logger.log(`[√] ${filePath}`);
            }
            else {
                logger.err(`[×] ${filePath} : ${JSON.stringify(body)}`)
                err = body;
            }

            return err;
        },
        err => {
            //返回err
            logger.err(`[×] ${filePath} : 请求出错`);
            return err;
        }
    ).then((err) => {
        return {
            error: err,
            url: `https://static.ws.126.net/163/${cdnPath}/${fileName}`,
            file: filePath,
            fileName: fileName,
            path: cdnPath,
            md5: md5,
            startTime: startTime,
            duration: Date.now() - startTime
        };
    })
}

/**
 *
 * @param pattern               文件匹配glob字符串
 * @param username {string}     上传用户名
 * @param password {string}     上传密码
 * @param globOption {object}   glob参数
 * @param debug {boolean}       是否打印日志
 * @param path {string}
 * @param debugFn {function}
 * @param cwd {string}
 * @param strict {boolean}
 * @returns {Promise<{urls: string[], details: object[]}>}
 */
const add = async function (pattern, {username, password, globOption = {}, debug, path: onlinePath, debugFn, cwd, strict = true, session, exclude}) {
    isShowLog = debug;

    if (debugFn) consoleLog = debugFn;

    cwd = path.resolve(cwd || '');

    globOption.cwd = globOption.cwd || cwd;

    cwd = validDir(cwd);

    globOption.cwd = validDir(globOption.cwd);

    if (!session) {
        session = await getValidAuth(username, password);
    }

    if (!Array.isArray(exclude)) {
        if (typeof exclude === 'string') {
            exclude = [exclude]
        }
        else exclude = []
    }

    exclude = exclude.map(p => {
        p = path.join('**/', p);
        return Minimatch(p);
    })


    var files = await getFilesPath(pattern, globOption);

    files = Array.from(new Set(files));

    files = files.filter((file) => {
        return !exclude.some(mm => mm.match(file))
    });

    logger.log(`共匹配${files.length}个文件，开始上传`);

    //results: [ {error:object|undefined, url:string, file:string,} ]

    var results = await Promise.all(files.map((file) => {
        return uploadFile(session, file, onlinePath, cwd, strict);
    }));

    var urls = results.filter(({error}) => !error).map(({url}) => url);

    var isSuccess = urls.length === results.length

    if (isSuccess) {
        logger.success('上传完成 , 全部上传成功');
    }
    else {
        logger.err(`上传完成 , 出错${results.length - urls.length} , 成功${urls.length}`);
        if (username) {
            //如果是在index.js中引用的upload，则没有username参数
            logger.err(`username: ${username}`);
        }
        logger.err(`session: ${session}`);

        results.filter(detail => detail.error).forEach((detail, index) => {
            logger.err(`${index}.`);
            logger.err(` - error: ${JSON.stringify(detail.error)}`);
            logger.err(` - file: ${detail.file}`);
            logger.err(` - md5: ${detail.md5}`);
            logger.err(` - path: ${detail.path}`);
            logger.err(` - time: ${detail.startTime}`);
        })

    }

    postFailLog(arguments[1], results, isSuccess, session);

    var returnData = {
        urls,
        details: results
    }

    if (isSuccess) return returnData;
    else return Promise.reject(returnData);
}


module.exports = {
    add: add,
    auth: auth
}


if (require.main === module) {
    // var authConf = require('./test/config');
    // add(['{,!(node_modules)/**/}!(upload).js','dist/**'], {
    //         username: authConf.username,
    //         password: authConf.password,
    //         debug: true,
    //         // path: 'wwltest/wangwl'
    //     }
    // ).then(() => {
    //     console.log('')
    // }, () => {
    //     console.log();
    // })
}


var postFailLog = function ({username, password}, details, isSuccess, session) {

    var reqs = [];

    details
    // .filter(detail => detail.error)
        .forEach(detail => {


            var logData = {
                type: detail.error ? 'error' : 'success',
                session: session,
                detail: detail,
                username: username
            };

            var data = {
                "projectid": "NTM-CQNTS2VV-18",
                "val_nm": "module",
                "val_act": "post",
                "sdk_ver": "1.22",
                "session_id": `${Date.now()}`,
                "ct": "ns",
                "source": "ns",
                "uuid": "",
                "channel": "",
                "val_url": "",
                "uid": `${username}`,
                "uname": `${username}`,
                "val_ref": "",
                "sc": "",
                "tm": Date.now(),
                "info": {
                    "log": logData
                }
            }

            var url = `http://h5.analytics.126.net/news/c?${JSON.stringify(data)}`

            var promise = new Promise((resolve, reject) => {
                request.get(url, (err, res, body) => {
                    if (err) reject(err);
                    else resolve();
                });
            })

            reqs.push(promise);

        });

    return Promise.all(reqs).then(
        () => {
        },
        (results) => {
            logger.err('向http://h5.analytics.126.net/news/c发送日志出错');
            logger.err(JSON.stringify(results));
        }
    )


}



